
# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2023-03-10T22:42:59+0000",
 "dirty": false,
 "error": null,
 "full-revisionid": "587734f3b1d4f22d1ff5dd5f6f7fa3d18bab8d9b",
 "version": "0.5.3.post0.dev13"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
