:root {
    --pst-header-height: 80px;

    /* Colors they are being added inside a `rgb()` tag that's why they look like this 
    also this pattern is being used by the theme under the hood, it's not because I really
    like to use this rather weird pattern :) 
    */
    --pst-color-link: var(--pst-color-text-base);
    --pst-color-primary: 34, 37, 42;
    --pst-color-text-base: var(--pst-color-primary);
    --pst-color-preformatted-background: 245, 246, 251;
    --pst-color-inline-code: var(--text-blue);
    --pst-color-sidebar-link-active: var(--text-blue);
    --pst-color-sidebar-link-hover: var(--text-blue);
    --pst-color-toc-link-active: var(--pst-color-text-base);
    --pst-color-toc-link-hover: var(--blue);
    --pst-color-link-hover: var(--text-blue);
    --pst-color-toc-link: 144, 142, 137;
    --pst-color-navbar-link: 255, 255, 255;
    --pst-color-navbar-link-hover: var(--light-blue);
    --pst-color-admonition-note: 166, 227, 255;
    --border: 1px solid rgba(204, 206, 210);
    --blue: 31, 90, 255;
    --text-blue: 24, 72, 205;
    --light-blue: 76, 124, 255;
    --red: 217, 62, 0;
    --transition: all .5s ease-in-out;
    --pst-font-family-base: Inter;
    --pst-font-family-heading: Inter;
    --pst-font-family-monospace: Roboto Mono, monospace;
    --text-grey: 117, 117, 117;
}


body {
    font-size: 16px;
}


a:hover {
    text-underline-offset: 3px;
}

.bd-search {
    margin-top: 1.5rem;
}

.admonition .admonition-title {
    background-color: #fff !important;
    font-size: 1.1rem;
    font-weight: inherit;
    padding: 1.5rem .6rem .4rem 2.5rem

}

.admonition {
    background-color: #fff;
    border: var(--border) !important;
    box-shadow: none;
    margin: 2rem 0rem !important;
    padding-bottom: 2rem !important;
}

.admonition.note .admonition-title:before {
    padding-top:.2rem;
    content: url('note-icon.svg');
}

.admonition.tip .admonition-title:before, .admonition.seealso .admonition-title:before {
    padding-top:.2rem;
    content: url('tip-icon.svg');
}

.admonition.attention .admonition-title:before, .admonition.warning .admonition-title:before {
    padding-top:.2rem;
    content: url('warning-icon.svg');
}

.versionchanged {
    border: var(--border) !important;
    background-color: #fff !important;
    box-shadow: none;
    margin: 2rem 0rem !important;
}

.versionmodified.changed:before {
    color: rgb(var(--light-blue)) !important;
}

.shadow {
    box-shadow: none !important;
}

.card {
    border-radius: 10px;
    border: var(--border);
}

.card-header {
    border-radius: calc(10px - 1px) calc(10px - 1px) 0 0 !important;
    border-bottom: var(--border);
    text-transform: uppercase;
    font-weight: inherit;
    font-size: inherit;
}

.card-footer {
    border-radius: 0 0 calc(10px - 1px) calc(10px - 1px) !important;
}

:root {
    font-variant-ligatures: none;
}

pre {
    border: none;
    box-shadow: none;
}

.logo {
    padding-left: 4px;
    padding-top: 4px;
    padding-bottom: 4px;
}

.navbar-light .navbar-toggler {
    color: white;
    border-color: rgb(204, 206, 210);
    background-color:white;
    border-radius: 50%;
    width: 50px;
    height: 50px;
    margin-right: 1rem;
}

.hs-button, input.hs-button.primary-large {
    display: inline-block;
    font-weight: 400;
    text-align: center;
    vertical-align: middle;
    user-select: none;
    border: 1px solid transparent !important;
    padding: .375rem .75rem;
    font-size: large;
    line-height: 1.5;
    border-radius: .25rem;
    color: rgb(var(--pst-color-text-base));
    background-color: rgb(var(--light-blue));
    text-decoration: none !important;
    transition: var(--transition);
    margin-top: 1rem;
}

.hs-field-desc {
    font-size: inherit;
}

.hs-error-msg {
    color: rgb(var(--red));
}

.hbspt-form {
    margin: 3rem 1rem;
}

.no-list {
    list-style: none;
    padding-left: 0;
}

.hs-input {
    border: none;
    border-bottom: var(--border);
}

.hs-input:hover, .hs-input:focus, .hs-input:active {
    border-bottom-width: 2px;
    border-color: rgb(var(--pst-color-text-base));

}

legend.hs-field-desc {
    font-size: small;
    color: rgb(var(--text-grey));
}

select.hs-input, select.hs-input>option:first-child {
    color: rgb(var(--text-grey));
}

.hs_recaptcha {
    margin: 1rem 0rem;
}

select.hs-input>option {
    color: rgb(var(--pst-color-text-base))
}

input[type=file] {
    border: none;
}

fieldset.form-columns-1, fieldset.form-columns-2 {
    margin-top: 1rem !important;
}

.btn-full {
    color: rgb(var(--pst-color-text-base));
    background-color: rgb(var(--blue));
    text-decoration: none !important;
    color: white;
    transition: var(--transition);
}

.btn-full:hover, .btn-full:active, input.hs-button.primary.large:hover, input.hs-button.primary.large:active {
    background-color: rgb(var(--light-blue));
    color: white !important;
    transition: var(--transition);
}

a.sphinx-bs.btn.text-wrap.btn-full.btn-block.stretched-link.reference.external:hover,
a.sphinx-bs.btn.text-wrap.btn-full.btn-block.stretched-link.reference.internal:hover,
a.sphinx-bs.btn.text-wrap.btn-full.btn-block.reference.internal:hover {
    color: rgb(var(--pst-color-text-base));
    transition: var(--transition);
}

.toc-entry>.nav-link.active {
    font-weight: inherit;
}


.bd-sidebar {
    border-right: none;
}

.navbar-light {
    background: rgb(25, 27, 31) !important;
    box-shadow: none;
}

.navbar-light .navbar-nav>.active>.nav-link {
    color: white !important;
    border-bottom: 1.5px solid rgb(var(--blue));
    border-width: 90%;
    font-weight: inherit;
}

.navbar-light .navbar-nav .nav-link:focus {
    color: rgb(var(--blue));
}

.nav-item {
    padding-right: 1rem;
}


ul.navbar-nav:hover li {
    opacity: .7;
    transition: var(--transition);
}

ul.navbar-nav li.nav-item:hover {
    opacity: 1 !important;
    transition: var(--transition);
}

ul.current.nav.bd-sidenav li.toctree-l1.current.active a.current.reference.internal {
  font-weight: inherit;
}


#coiled-docs .section .reference>code {
    text-decoration: none !important;
    transition: var(--transition);
}

main a.reference.external,
a>span.doc,
a>span.std.std-ref {
    text-underline-offset: 3px;
    text-decoration: underline;
    text-decoration-thickness: 1.5px;
    text-decoration-color: rgb(var(--blue));
    transition: var(--transition);
}

#coiled-docs .reference:hover, .section .reference:hover {
    transition: var(--transition);
}

#coiled-docs .btn-full {
    text-decoration: none;
}

.btn-full>span.doc {
    text-decoration: none;
}

ul {
    list-style-type: circle;
}

li::marker {
    font-size: 21px;
}

.tocsection {
    border-left: none !important;
}

.bd-toc .onthispage {
    color: rgb(var(--pst-color-text-base));
}

.fa-angle-left:before {
    content: url("left-arrow.svg");
}

.fa-angle-right:before {
    content: url("right-arrow.svg");
}

.left-prev, .right-next {
    border: none !important;
    color: rgb(var(--pst-color-text-base)) !important;
    text-decoration: none !important;
}

.left-prev:hover, .left-prev:focus, .right-next:hover, .right-next:focus {
    text-underline-offset: 3px;
    text-decoration: underline !important;
}

.table thead th {
    text-transform: uppercase;
    font-weight: inherit;
}

.table td, .table th {
    border-top: var(--border);
}

tr {
    border-bottom: var(--border);
}

.copybtn {
    cursor: pointer;
    margin-top: .5rem;
    margin-right: .5rem;
}

i.fas.fa-list, i.fas.fa-external-link-alt {
    display: none;
}

.tabbed-content {
    box-shadow: none;
}

.tabbed-set>label {
    border-bottom: 2px solid rgb(210, 208, 204);
    font-weight:inherit;
}

/* Code highlight */

.highlight .kn, .highlight .nb, .highlight .k, .highlight .mi, .highlight .kc, code.xref, a code {
    color: rgb(var(--pst-color-inline-code));
    font-weight: inherit;
    text-decoration: none !important;
}


.highlight .kn, .highlight .o, .highlight .nb {
    color: rgb(var(--text-blue));
    font-weight: 400;
}

.highlight .nt {
    font-weight: 400;
}

.highlight .c1 {
    color: rgb(105, 105, 105);
}

.highlight .s2 {
    color: rgb(var(--text-blue));
}

.highlight .ne {
    color: rgb(var(--text-blue));
}


/* Media query for small screens */
@media (max-width: 958px) {
    .navbar-light .navbar-nav>.active>.nav-link {
        margin-right: 80%;
    }
    ul#navbar-main-elements {
        margin: 1rem 0rem 0rem 2rem;
    }
    .nav-item {
        margin-top: 1rem;
    }

    ul#navbar-icon-links {
        flex-direction: row;
        margin-left: 2rem;
        padding-bottom: 1rem;
    }
    .navbar-brand {
        margin-left: .5rem;
    }
}

@media (max-width: 675px) {
    .navbar-light .navbar-nav>.active>.nav-link {
        margin-right: 68%;
    }
}

@media (min-width: 750px) {
    .bd-links {
        max-height: calc(100vh - 15rem) !important;
    }
}

/* sphinx tabs customization */
.sphinx-tabs-tab {
    font-family: var(--pst-font-family-base);
    color: var(--border);
    font-size: 1.1rem;
    background-color: rgba(255, 255, 255, 0);
    padding: 1rem 1.5rem;
    margin-bottom: 0;
}

.sphinx-tabs-tab[aria-selected="true"] {
    border: var(--border);
}

.sphinx-tabs-panel {
    border: var(--border);
}

[role="tablist"] {
    border-bottom: var(--border);
}